!*************************************************!
! Qun Liu 2015-3-22
!*************************************************!

program  main
    use mpi
    implicit none
    integer ierr

    call MPI_INIT(ierr)
    call printInfo
    call MPI_BARRIER(MPI_COMM_WORLD, ierr)
    call dataTransfer
    call MPI_BARRIER(MPI_COMM_WORLD, ierr)
    call sayHello
    call MPI_FINALIZE(ierr) 
    stop
end program main

subroutine printInfo
    use mpi
    implicit none
    integer rank, size, ierr

    call MPI_COMM_RANK(MPI_COMM_WORLD, rank, ierr) 
    call MPI_COMM_SIZE(MPI_COMM_WORLD, size, ierr)
    
    write(*,10) rank, size 
    10 format("Process ID is ", I4, ", the number of total processes is ", I4, ".")
    return
end subroutine printInfo

subroutine dataTransfer
    use mpi
    implicit none
    integer :: rank, size, ierr, senddata, recvdata, tag=1
    integer, dimension(MPI_STATUS_SIZE) :: status

    call MPI_COMM_RANK(MPI_COMM_WORLD, rank, ierr) 
    call MPI_COMM_SIZE(MPI_COMM_WORLD, size, ierr)
    
    if (size<2) then
        write(*,*) "Error: the num of processes should >= 2, please try again."
        return 
    end if

    if (rank==0) then
        write(*,*) "Please input a data(integer) to transfer:"
        read(*,*),senddata

        call MPI_SEND(senddata, 1, MPI_INTEGER, rank+1, tag, MPI_COMM_WORLD, ierr)
        write(*,10) rank, senddata, rank+1
    else if (rank>0 .and. rank<size-1) then
        call MPI_RECV(recvdata, 1, MPI_INTEGER, rank-1, tag, MPI_COMM_WORLD, status, ierr)
        write(*,11) rank, recvdata, rank-1
        call MPI_SEND(recvdata, 1, MPI_INTEGER, rank+1, tag, MPI_COMM_WORLD, ierr)
        write(*,10) rank, recvdata, rank+1
    else
        call MPI_RECV(recvdata, 1, MPI_INTEGER, rank-1, tag, MPI_COMM_WORLD, status, ierr)
        write(*,11) rank, recvdata, rank-1
    end if

    10 format("Process ", I4, " sends    data ", I5, " to   process ", I4, ".")
    11 format("Process ", I4, " receives data ", I5, " from process ", I4, ".")

    return
end subroutine dataTransfer

subroutine sayHello
    use mpi
    implicit none
    integer         :: i, rank, size, ierr, tag=2
    character*5     :: senddata="hello", recvdata
    integer, dimension(MPI_STATUS_SIZE) :: status

    call MPI_COMM_RANK(MPI_COMM_WORLD, rank, ierr) 
    call MPI_COMM_SIZE(MPI_COMM_WORLD, size, ierr)
    
    if (size<2) then
        write(*,*) "Error: the num of processes should >= 2, please try again."
        return 
    end if

    do i=0,size-1
        if(i .ne. rank) then
            call MPI_SEND(senddata, len(senddata), MPI_CHARACTER, i, tag, MPI_COMM_WORLD, ierr)
            write(*,10) rank, senddata, i 
            call MPI_RECV(recvdata, len(senddata), MPI_CHARACTER, MPI_ANY_SOURCE, tag, MPI_COMM_WORLD, status, ierr)
            write(*,11) rank, recvdata, status(MPI_SOURCE)
        end if  
    end do

    10 format("Process ", I4, " send     message """, A5, """ to   process ", I4, ".")
    11 format("Process ", I4, " receives message """, A5, """ from process ", I4, ".")
    
    return
end subroutine sayHello
